import os
from .utils import load_lib

class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Also, the decorated class cannot be
    inherited from. Other than that, there are no restrictions that apply
    to the decorated class.

    To get the singleton instance, use the `instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    """

    def __init__(self, decorated):
        self._decorated = decorated

    def instance(self):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.

        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)


@Singleton
class LibVorbis:
    def __init__(self):
        oldDir = os.getcwd()
        os.chdir("lib/pythonfsb5")
        self._lib = load_lib("vorbis")
        os.chdir(oldDir)

    def getLib(self):
        return self._lib


@Singleton
class LibOgg:
    def __init__(self):
        oldDir = os.getcwd()
        os.chdir("lib/pythonfsb5")
        self._lib = load_lib("ogg")
        os.chdir(oldDir)

    def getLib(self):
        return self._lib
